from .checks import *
from .plot import *
from .utils import *
from .._basics._base_objects import _FairnessObject
from ..._theme import get_default_config


class GroupFairnessClassification(_FairnessObject):

    def __init__(self, y, y_hat, protected, privileged, label, verbose=False, cutoff=0.5):

        super().__init__(y, y_hat, protected, privileged, verbose)

        cutoff = check_cutoff(self.protected, cutoff, verbose)
        self.cutoff = cutoff

        sub_confusion_matrix = SubgroupConfusionMatrix(y_true=self.y,
                                                       y_pred=self.y_hat,
                                                       protected=self.protected,
                                                       cutoff=self.cutoff)

        sub_confusion_matrix_metrics = SubgroupConfusionMatrixMetrics(sub_confusion_matrix)
        df_ratios = calculate_ratio(sub_confusion_matrix_metrics, self.privileged)
        parity_loss = calculate_parity_loss(sub_confusion_matrix_metrics, self.privileged)

        self._subgroup_confusion_matrix = sub_confusion_matrix
        self._subgroup_confusion_matrix_metrics_object = sub_confusion_matrix_metrics
        self.metric_scores = sub_confusion_matrix_metrics.to_horizontal_DataFrame()
        self.parity_loss = parity_loss
        self.result = df_ratios
        self.label = label

    def fairness_check(self, epsilon=0.8, verbose=True):
        """Check if classifier passes popular fairness metrics

        Fairness check is easy way to check if model is fair.
        For that method uses 5 popular metrics of group fairness.
        Model is considered to be fair if confusion matrix
        metrics are close to each other.
        This arbitrary decision is based on epsilon, which default value
        is 0.8 which matches four-fifths (80%) rule.

        Methods in use: Equal opportunity, Predictive parity, Predictive equality,
        Statistical parity and Accuracy equality.

        Parameters
        -----------
        epsilon : float (default 0.8)
            Parameter defines acceptable fairness scores. The closer to 1 the
            more strict the verdict is. If the ratio of certain unprivileged
            and privileged subgroup is within (epsilon, 1/epsilon) than
            there is no discrimination in this metric and for this subgroups.
        verbose : boolean
            Shows communicate about potential problems (NAN's in model metrics)
            that can cause misinterpretation.

        Returns
        -----------
        Console output

        """
        epsilon = check_epsilon(epsilon)
        metric_ratios = self.result

        subgroups = np.unique(self.protected)
        subgroups_without_privileged = subgroups[subgroups != self.privileged]
        metric_ratios = metric_ratios.loc[subgroups_without_privileged, fairness_check_metrics()]

        metrics_exceeded = ((metric_ratios > 1 / epsilon) | (epsilon > metric_ratios)).apply(sum, 0)

        names_of_exceeded_metrics = list(metrics_exceeded.index[metrics_exceeded != 0])
        if len(names_of_exceeded_metrics) >= 2:
            print(f'Bias detected in {len(names_of_exceeded_metrics)} metrics: {", ".join(names_of_exceeded_metrics)}')
        elif len(names_of_exceeded_metrics) == 1:
            print(f'Bias detected in {len(names_of_exceeded_metrics)} metric: {names_of_exceeded_metrics[0]}')
        else:
            print("No bias was detected!")

        # arbitrary decision
        if len(names_of_exceeded_metrics) >= 2:
            conclusion = 'is not fair because 2 or more metric scores exceeded acceptable limits set by epsilon'
        elif len(names_of_exceeded_metrics) == 1:
            conclusion = 'cannot be called fair because 1 metric score exceeded acceptable limits set by epsilon.\n' \
                         'It does not mean that your model is unfair ' \
                         'but it cannot be automatically approved based on these metrics'
        else:
            conclusion = 'is fair in terms of checked fairness metrics'

        print(f'\nConclusion: your model {conclusion}.')

        print(
            f'\nRatios of metrics, based on \'{self.privileged}\'. Parameter \'epsilon\' was set to {epsilon}'
            f' and therefore metrics should be within ({epsilon}, {round(1 / epsilon, 3)})')
        print(metric_ratios.to_string())
        if np.isnan(metric_ratios).sum().sum() > 0:
            verbose_cat(
                '\nWarning!\nTake into consideration that NaN\'s are present, consider checking \'metric_scores\' '
                'plot to see the difference', verbose=verbose)

        return

    def plot(self,
             objects=None,
             type='fairness_check',
             title=None,
             show=True,
             **kwargs):
        """
        Parameters
        -----------
        objects : GroupFairnessClassification object
            Additional objects to plot (default is None).
        type : str, optional
            Type of the plot. Default is 'fairness_check'.
            When the type of plot is specified, user may provide additional
            keyword arguments (**kwargs) which will
            be used in creating plot of certain type.
            Below there is list of types and **kwargs used by them

            fairness_check:
                fairness_check plot visualizes the fairness_check method
                for one or more GroupFairnessClassification objects.
                It accepts following keyword arguments:
                 'epsilon' - which denotes the decision
                             boundary (like in fairness_check method)

            metric_scores:
                metric_scores plot shows real values of metrics.
                Each model displays values in each metric and each subgroup.
                Vertical lines show metric score for privileged
                subgroup and points connected with the lines
                show scores for unprivileged subgroups.
                This plot is simple and it does
                not have additional keyword arguments.

            stacked:
                stacked plot shows cumulated parity loss from chosen
                metrics. It stacks metrics on top of each other.
                It accepts following keyword arguments:
                'metrics' - list of metrics to be plotted. The metrics are taken
                            from parity_loss attribute of the object.
                            Default is ["TPR", "ACC", "PPV", "FPR", "STP"].

            radar:
                radar plot shows parity loss of provided metrics. It does it
                in form of radar (spider) chart. The smaller the field of
                figure the better.
                It accepts following keyword arguments:
                'metrics' - list of metrics to be plotted. The metrics are taken
                            from parity_loss attribute of the object.
                            Default is ["TPR", "ACC", "PPV", "FPR", "STP"].

            performance_and_fairness:
                performance_and_fairness plot shows relation between chosen
                performance and fairness metrics. The fairness metric axis is
                reversed, because the higher the model the less bias it has.
                Thanks to that it is more intuitive to look at because
                the best models are in top right corner.
                It accepts following keyword arguments:
                'fairness_metric' - single fairness metric to be plotted on Y axis.
                                   The metric is taken from parity_loss attribute\
                                   of the object. The default is "TPR"
                'performance_metric' - single performance metric. One of {'recall',
                                       'precision','accuracy','auc','f1'}.
                                       Metrics apart from 'auc' are
                                       cutoff-sensitive. Default is "accuracy"


            heatmap:
                heatmap shows parity loss of metrics in form of heatmap. The less
                parity loss model has, the more fair it is.
                It accepts following keyword arguments:
                'metrics' - list of metrics to be plotted. The metrics are taken
                            from parity_loss attribute of the object.
                            Default is 'all' which stands for all available metrics.

            ceteris_paribus_cutoff:
                ceteris_paribus_cutoff plot shows what would happen if cutoff
                for only one subgroup would change with others cutoffs constant.
                The plot shows also a minimum, where sum of parity loss of metrics
                is the lowest. Minimum only works if at some interval all metrics
                have non-nan scores.
                It accepts following keyword arguments:
                'subgroup' - necessary argument. It is name of subgroup from
                             protected attribute. Cutoff for this subgroup will
                             be changed.

                'metrics' - list of metrics to be plotted. The metrics are taken
                            from parity_loss attribute of the object.
                            Default is ["TPR", "ACC", "PPV", "FPR", "STP"].

                'grid_points' - number of grid points (cutoff values) to calculate
                                metrics for. The points are distributed evenly.
                                Default is 101.





        title : str, optional
            Title of the plot (default depends on the `type` attribute).
        show : bool, optional
            True shows the plot; False returns the plotly Figure object that can be
            edited or saved using the `write_image()` method (default is True).

        Returns
        -----------
        None or plotly.graph_objects.Figure
            Return figure that can be edited or saved. See `show` parameter.
        """
        other_objects = None
        if objects is not None:
            other_objects = []
            for obj in objects:
                if isinstance(obj, self.__class__):
                    other_objects.append(obj)
            check_other_fairness_objects(self, other_objects)

        if type == 'fairness_check':
            fig = plot_fairness_check(self,
                                      other_objects=other_objects,
                                      title=title, **kwargs)

        elif type == "metric_scores":
            fig = plot_metric_scores(self,
                                     other_objects=other_objects,
                                     title=title,
                                     **kwargs)

        # names of plots may be changed
        elif type == 'stacked':
            fig = plot_stacked(self,
                               other_objects=other_objects,
                               title=title,
                               **kwargs)

        elif type == 'radar':
            fig = plot_radar(self,
                             other_objects=other_objects,
                             title=title,
                             **kwargs)

        elif type == 'performance_and_fairness':
            fig = plot_performance_and_fairness(self,
                                                other_objects=other_objects,
                                                title=title,
                                                **kwargs)

        elif type == 'heatmap':
            fig = plot_heatmap(self,
                               other_objects=other_objects,
                               title=title,
                               **kwargs)

        elif type == 'ceteris_paribus_cutoff':
            fig = plot_ceteris_paribus_cutoff(self,
                                              other_objects=other_objects,
                                              title=title,
                                              **kwargs)

        else:
            raise ParameterCheckError(f"plot type {type} not supported, try other types.")

        if show:
            fig.show(config=get_default_config())
        else:
            return fig



import numpy as np
import plotly.graph_objects as go


def get_default_colors(n, type):
    default_colors = ["#8bdcbe", "#f05a71", "#371ea3", "#46bac2", "#ae2c87", "#ffa58c", "#4378bf"]

    if n > len(default_colors):
        ret = []
        for _ in range(np.ceil(n / len(default_colors)).astype(int)):
            ret += default_colors
        return ret

    bar_colors = {
        1: ["#46bac2"],
        2: ["#8bdcbe", "#4378bf"],
        3: ["#8bdcbe", "#4378bf", "#46bac2"],
        4: ["#46bac2", "#371ea3", "#8bdcbe", "#4378bf"],
        5: ["#8bdcbe", "#f05a71", "#371ea3", "#46bac2", "#ffa58c"],
        6: ["#8bdcbe", "#f05a71", "#371ea3", "#46bac2", "#ae2c87", "#ffa58c"],
        7: default_colors
    }

    line_colors = {
        1: ["#46bac2"],
        2: ["#8bdcbe", "#4378bf"],
        3: ["#8bdcbe", "#f05a71", "#4378bf"],
        4: ["#8bdcbe", "#f05a71", "#4378bf", "#ffa58c"],
        5: ["#8bdcbe", "#f05a71", "#4378bf", "#ae2c87", "#ffa58c"],
        6: ["#8bdcbe", "#f05a71", "#46bac2", "#ae2c87", "#ffa58c", "#4378bf"],
        7: default_colors
    }

    if type == 'bar':
        return bar_colors[n]
    elif type == 'line':
        return line_colors[n]
    else:
        return bar_colors[n]


def get_break_down_colors():
    return ["#371ea3", "#8bdcbe", "#f05a71"]


def get_default_config():
    return {
                'displaylogo': False, 'staticPlot': False,
                'toImageButtonOptions': {'height': None, 'width': None, },
                'modeBarButtonsToRemove': ['sendDataToCloud', 'lasso2d', 'autoScale2d', 'select2d', 'zoom2d', 'pan2d',
                                           'zoomIn2d', 'zoomOut2d', 'resetScale2d', 'toggleSpikelines',
                                           'hoverCompareCartesian', 'hoverClosestCartesian']
            }


def fig_update_line_plot(fig, title, y_title, plot_height, hovermode):

    for axis in fig.layout:
        if type(fig.layout[axis]) == go.layout.YAxis:  # remove redundant axis labels
            fig.layout[axis].title.text = ''
        elif type(fig.layout[axis]) == go.layout.XAxis:  # remove redundant axis labels
            fig.layout[axis].title.text = ''
        elif axis == 'annotations':
            for index, annotation in enumerate(fig.layout[axis]):  # fix annotation text
                annotation.update(text=annotation.text.split("=")[-1],
                                  font=dict(size=13))  # , x=0, xref='x' + str(index+1))  # title on the left bug

    fig.update_layout(
        # keep the original annotations and add axis title
        annotations=list(fig.layout.annotations) + [
            go.layout.Annotation(
                x=-0.07,
                y=0.5,
                font=dict(size=13),
                showarrow=False,
                text=y_title,
                textangle=-90,
                xref="paper",
                yref="paper"
            )
        ],
        font=dict(color="#371ea3"),
        margin=dict(t=78, b=71, r=30),
        hovermode=hovermode,
        title=dict(text=title, x=0.15, font=dict(size=16)),  # y=1 - 50/plot_height,
        legend=dict(
            title=dict(font=dict(size=12)),
            orientation="h",
            yanchor="bottom",
            y=1 + 30 / plot_height,
            xanchor="right",
            x=1,
            itemsizing='constant',
            font=dict(size=11)
        ),
        height=plot_height
    )

    return fig



import numpy as np
from ..._explainer.helper import verbose_cat
from .._basics.exceptions import ParameterCheckError


def check_cutoff(protected, cutoff, verbose):
    if isinstance(cutoff, dict):
        for value in cutoff.values():
            if not isinstance(value, float):
                raise ParameterCheckError("cutoff values must be floats")
            else:
                for value in cutoff.values():
                    if not 0 < value < 1:
                        raise ParameterCheckError("cutoff values should be in [0, 1] range")

        # changing to sets
        cutoff_set = set(cutoff.keys())
        protected_set = set(protected)

        # all cutoff keys are in protected
        if cutoff_set | protected_set != protected_set:
            raise ParameterCheckError("cutoff dict contains keys not present in protected")

        # if some keys are in protected but not in cutoff, add them to cutoff with default value
        if protected_set - cutoff_set != set():
            verbose_cat("Adding default (0.5) cutoffs for subgroups not provided in cutoff keys", verbose)
            for subgroup in protected_set - cutoff_set:
                cutoff[subgroup] = 0.5

    elif isinstance(cutoff, float):
        if not 0 < cutoff < 1:
            raise ParameterCheckError("if cutoff is a float it should be in [0, 1] range")
        verbose_cat("Converting cutoff to dict", verbose)
        subgroups = np.unique(protected)
        value = cutoff
        cutoff = {}
        for subgroup in subgroups:
            cutoff[subgroup] = value

    else:
        raise ParameterCheckError(f"cutoff must be a float or a dict, not {type(cutoff)}")

    return cutoff


def check_epsilon(epsilon):
    if not isinstance(epsilon, float):
        raise ParameterCheckError("epsilon must be float")

    if not 0 < epsilon < 1:
        raise ParameterCheckError("epsilon must be in (0,1) range")

    return epsilon




import unittest

import numpy as np
from sklearn.compose import ColumnTransformer
from sklearn.impute import SimpleImputer
from sklearn.neural_network import MLPRegressor
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import StandardScaler, LabelEncoder, OneHotEncoder

import dalex as dx


class PredictTestTitanic(unittest.TestCase):
    def setUp(self):
        data = dx.datasets.load_titanic()
        data.loc[:, 'survived'] = LabelEncoder().fit_transform(data.survived)

        self.X = data.drop(columns='survived')
        self.y = data.survived

        numeric_features = ['age', 'fare', 'sibsp', 'parch']
        numeric_transformer = Pipeline(steps=[
            ('imputer', SimpleImputer(strategy='median')),
            ('scaler', StandardScaler())])

        categorical_features = ['gender', 'class', 'embarked']
        categorical_transformer = Pipeline(steps=[
            ('imputer', SimpleImputer(strategy='constant', fill_value='missing')),
            ('onehot', OneHotEncoder(handle_unknown='ignore'))])

        preprocessor = ColumnTransformer(
            transformers=[
                ('num', numeric_transformer, numeric_features),
                ('cat', categorical_transformer, categorical_features)])

        clf = Pipeline(steps=[('preprocessor', preprocessor),
                              ('classifier', MLPRegressor(hidden_layer_sizes=(150, 100, 50),
                                                          max_iter=500, random_state=0))])

        clf.fit(self.X, self.y)

        self.exp = dx.Explainer(clf, self.X, self.y, verbose=False)

    def test(self):
        self.assertIsInstance(self.exp.predict(self.X.iloc[[0]]), np.ndarray)
        with self.assertRaises(TypeError):
            self.exp.predict(self.X.iloc[0])
        with self.assertRaises(ValueError):
            self.exp.predict(self.X.iloc[0].values)

        self.assertIsInstance(self.exp.predict(self.X.iloc[:100]), np.ndarray)


if __name__ == '__main__':
    unittest.main()



# dx.Explainer
# from dalex.datasets import load_*
from . import datasets
from ._explainer.object import Explainer
from ._arena.object import Arena

__version__ = '0.4.1'

__all__ = [
    "Explainer",
    "dataset_level",
    "instance_level",
    "fairness",
    "datasets",
    "Arena"
]



import unittest

import dalex as dx
import pandas as pd


class DatasetsTest(unittest.TestCase):
    def setUp(self):
        pass

    def test(self):
        case1 = dx.datasets.load_fifa()
        case2 = dx.datasets.load_titanic()
        case3 = dx.datasets.load_apartments()
        case4 = dx.datasets.load_apartments_test()
        case5 = dx.datasets.load_hr()
        case6 = dx.datasets.load_hr_test()
        case7 = dx.datasets.load_dragons()
        case8 = dx.datasets.load_dragons_test()

        self.assertIsInstance(case1, pd.DataFrame)
        self.assertIsInstance(case2, pd.DataFrame)
        self.assertIsInstance(case3, pd.DataFrame)
        self.assertIsInstance(case4, pd.DataFrame)
        self.assertIsInstance(case5, pd.DataFrame)
        self.assertIsInstance(case6, pd.DataFrame)
        self.assertIsInstance(case7, pd.DataFrame)
        self.assertIsInstance(case8, pd.DataFrame)


if __name__ == '__main__':
    unittest.main()



import plotly.express as px
import plotly.graph_objects as go
from plotly.subplots import make_subplots

from .utils import *
from .utils import _metric_ratios_2_df, _unwrap_parity_loss_data, _fairness_theme, _classification_performance
from .._basics.checks import check_other_fairness_objects
from .._basics.exceptions import ParameterCheckError
from ..._explainer.helper import verbose_cat
from ..._theme import get_default_colors


def plot_fairness_check(fobject,
                        title=None,
                        other_objects=None,
                        epsilon=0.8,
                        verbose=True):
    data = _metric_ratios_2_df(fobject)
    n = 1
    if other_objects is not None:
        check_other_fairness_objects(fobject, other_objects)
        for other_obj in other_objects:
            other_data = _metric_ratios_2_df(other_obj)
            data = data.append(other_data)
            n += 1

    if any(data.score == 0):
        nan_models = set(data.label[data.score == 0])
        verbose_cat(f'\nFound NaN\'s or 0\'s for models: {nan_models}\n'
                    f'It is advisable to check \'metric_ratios\'', verbose=verbose)

    upper_bound = max([max(data.score[np.invert(np.isnan(data.score.to_numpy()))]), 1 / epsilon - 1]) + 0.1
    lower_bound = min([min(data.score[np.invert(np.isnan(data.score.to_numpy()))]), epsilon - 1]) - 0.1
    lower_bound = round(lower_bound, 1)
    upper_bound = round(upper_bound, 1)

    # including upper bound
    ticks = np.arange(lower_bound, upper_bound + 0.001, step=0.1).round(1)

    # drwhy colors
    colors = get_default_colors(n, 'line')

    # change name of metrics
    data.loc[data.metric == 'TPR', 'metric'] = 'Equal opportunity ratio     TP/(TP + FN)'
    data.loc[data.metric == 'ACC', 'metric'] = 'Accuracy equality ratio    (TP + TN)/(TP + FP + TN + FN)'
    data.loc[data.metric == 'PPV', 'metric'] = 'Predictive parity ratio     TP/(TP + FP)'
    data.loc[data.metric == 'FPR', 'metric'] = 'Predictive equality ratio   FP/(FP + TN)'
    data.loc[data.metric == 'STP', 'metric'] = 'Statistical parity ratio   (TP + FP)/(TP + FP + TN + FN)'

    # without privileged
    data = data.loc[data.subgroup != fobject.privileged]

    # subgroup y-axis value creation
    n_ticks = len(data.subgroup.unique())
    tick_values = np.linspace(0, 1, n_ticks + 2)[1:-1]

    subgroup_tick_dict = {}
    for i in range(len(data.subgroup.unique())):
        subgroup = data.subgroup.unique()[i]
        subgroup_tick_dict[subgroup] = tick_values[i]

    data['subgroup_numeric'] = [subgroup_tick_dict.get(sub) for sub in data.subgroup]
    data = data.reset_index(drop=True)

    # for hover
    data['dispx'] = np.round(data.score + 1, 3)
    # make fig
    fig = px.bar(data,
                 y='subgroup_numeric',
                 x='score',
                 color='label',
                 color_discrete_sequence=colors,
                 facet_col='metric',
                 facet_col_wrap=1,
                 barmode='group',
                 orientation='h',
                 custom_data=['subgroup', 'dispx', 'label']
                 )

    fig.update_traces(
        hovertemplate="<br>".join([
            "<b>%{customdata[2]}</b><br>"
            "Subgroup: %{customdata[0]}",
            "Score: %{customdata[1]}",
            "<extra></extra>"

        ]))

    fig.update_yaxes(tickvals=list(subgroup_tick_dict.values()),
                     ticktext=list(subgroup_tick_dict.keys()))

    # change axes range and labs
    fig.update_xaxes(tickvals=ticks,
                     ticktext=(ticks + 1).round(1),
                     range=[lower_bound, upper_bound])

    # refs are dependent on fixed numbers of metrics
    refs = ['y', 'y2', 'y3', 'y4', 'y5']
    left_red = [{'type': "rect",
                 'x0': lower_bound,
                 'y0': 0,
                 'x1': epsilon - 1,
                 'y1': 1,
                 'yref': yref,
                 'line': {'width': 0},
                 'fillcolor': '#f05a71',
                 'layer': 'below',
                 'opacity': 0.1} for yref in refs]

    middle_green = [{'type': "rect",
                     'x0': epsilon - 1,
                     'y0': 0,
                     'x1': 1 / epsilon - 1,
                     'y1': 1,
                     'yref': yref,
                     'line': {'width': 0},
                     'fillcolor': '#c7f5bf',
                     'layer': 'below',
                     'opacity': 0.1} for yref in refs]

    right_red = [{'type': "rect",
                  'x0': 1 / epsilon - 1,
                  'y0': 0,
                  'x1': upper_bound,
                  'y1': 1,
                  'yref': yref,
                  'line': {'width': 0},
                  'fillcolor': '#f05a71',
                  'layer': 'below',
                  'opacity': 0.1} for yref in refs]

    black_line = [{'type': 'line',
                   'x0': 0,
                   'x1': 0,
                   'y0': 0,
                   'y1': 1,
                   'yref': yref,
                   'xref': "x",
                   'line': {'color': "#371ea3", 'width': 1.5}} for yref in refs]

    for i in range(len(refs)):
        fig.add_shape(right_red[i])
        fig.add_shape(left_red[i])
        fig.add_shape(middle_green[i])
        fig.add_shape(black_line[i])

        fig.update_shapes(dict(xref='x'))

    if title is None:
        title = 'Fairness Check'

    fig.update_layout(_fairness_theme(title))

    # delete 'metric=' from facet names
    fig.for_each_annotation(
        lambda a: a.update(text=a.text.replace("metric=", ""), xanchor='left', x=0.05, font={'size': 15}))

    # delete y axis names [fixed] number of refs
    for i in ['', '2', '4', '5']:
        fig.update_layout({'yaxis' + i + '_title_text': ''})

    fig.update_layout({'yaxis3_title_text': 'subgroup'})
    fig.update_yaxes(showgrid=False, zeroline=False)

    return fig


def plot_metric_scores(fobject,
                       other_objects,
                       title=None):
    data = fobject._subgroup_confusion_matrix_metrics_object.to_vertical_DataFrame()
    data['label'] = np.repeat(fobject.label, data.shape[0]).astype('U')
    n = 1
    if other_objects is not None:
        check_other_fairness_objects(fobject, other_objects)
        for other_obj in other_objects:
            other_data = other_obj._subgroup_confusion_matrix_metrics_object.to_vertical_DataFrame()
            other_data['label'] = np.repeat(other_obj.label, other_data.shape[0]).astype('U')
            data = data.append(other_data)
            n += 1

    # metric choosing and name change
    data = data.loc[data.metric.isin(["TPR", "PPV", "STP", "ACC", "FPR"])]
    data.loc[data.metric == 'TPR', 'metric'] = 'TPR    TP/(TP + FN)'
    data.loc[data.metric == 'ACC', 'metric'] = 'ACC   (TP + TN)/(TP + FP + TN + FN)'
    data.loc[data.metric == 'PPV', 'metric'] = 'PPV    TP/(TP + FP)'
    data.loc[data.metric == 'FPR', 'metric'] = 'FPR    FP/(FP + TN)'
    data.loc[data.metric == 'STP', 'metric'] = 'STP   (TP + FP)/(TP + FP + TN + FN)'

    # for x axis
    min_score = min(data.score)
    max_score = max(data.score)

    # subgroup y-axis value creation
    n_ticks = len(data.label.unique())
    tick_values = np.linspace(0, 1, n_ticks + 2)[1:-1]

    privileged_data = data.loc[data.subgroup == fobject.privileged]
    data = data.loc[data.subgroup != fobject.privileged]

    subgroup_tick_dict = {}
    label_tick_dict = {}
    for i in range(len(data.subgroup.unique())):
        subgroup = data.subgroup.unique()[i]
        subgroup_tick_dict[subgroup] = i

    for i in range(len(data.label.unique())):
        label = data.label.unique()[i]
        label_tick_dict[label] = tick_values[i]

    data['subgroup_numeric'] = [subgroup_tick_dict.get(sub) for sub in data.subgroup]
    data = data.reset_index(drop=True)
    data.subgroup_numeric = data.subgroup_numeric + pd.Series([label_tick_dict.get(lab) for lab in data.label])

    # drwhy colors
    colors = get_default_colors(len(data.label.unique()), 'line')

    # fig creation
    fig = px.scatter(data,
                     x='score',
                     y='subgroup_numeric',
                     color='label',
                     color_discrete_sequence=colors,
                     facet_col='metric',
                     facet_col_wrap=1,
                     custom_data=['subgroup', 'label'])

    fig.update_traces(
        hovertemplate="<br>".join([
            "<b>%{customdata[1]}</b><br>"
            "Subgroup: %{customdata[0]}",
            "Score: %{x}"
            "<extra></extra>"

        ]))

    # point size
    fig.update_traces(mode='markers',
                      marker_size=7)

    # axis ticks
    fig.update_xaxes(tickvals=np.arange(0, 1.01, 0.1))

    # cols and ref dicts are dependent on the arrangement of metrics and labels
    color_dict = {}
    k = 0
    for label in data.label.unique():
        color_dict[label] = colors[k]
        k += 1

    refs = ['y', 'y2', 'y3', 'y4', 'y5']
    refs_dict = {}
    j = 1
    for metric in data.metric.unique():
        refs_dict[metric] = refs[len(refs) - j]
        j += 1
    # add lines
    for metric in data.metric.unique():
        for label in data.label.unique():
            x = float(privileged_data.loc[(privileged_data.metric == metric) &
                                          (privileged_data.label == label), :].score)
            # lines
            for subgroup in data.subgroup.unique():
                y = float(data.loc[(data.metric == metric) &
                                   (data.label == label) &
                                   (data.subgroup == subgroup)].subgroup_numeric)
                # horizontal
                fig.add_shape(type='line',
                              xref='x',
                              yref=refs_dict.get(metric),
                              x0=float(data.loc[(data.metric == metric) &
                                                (data.label == label) &
                                                (data.subgroup == subgroup)].score),
                              x1=x,
                              y0=y,
                              y1=y,
                              line=dict(
                                  color=color_dict.get(label),
                                  width=1))
            # vertical
            fig.add_shape(type='line',
                          xref='x',
                          yref=refs_dict.get(metric),
                          x0=x,
                          x1=x,
                          y0=0,
                          y1=np.ceil(max(data.subgroup_numeric)),
                          line=dict(
                              color=color_dict.get(label),
                              width=2))

    # theme and appearance
    if title is None:
        title = 'Metric Scores'

    fig.update_layout(_fairness_theme(title))

    fig.for_each_annotation(
        lambda a: a.update(text=a.text.replace("metric=", ""), xanchor='left', x=0.05, font={'size': 15}))

    # disable all y grids and
    # delete y axis names [fixed] number of refs
    for i in ['', '2', '3', '4', '5']:
        fig.update_layout({'yaxis' + i + '_showgrid': False,
                           'yaxis' + i + '_zeroline': False})
        if i != '3':
            fig.update_layout({'yaxis' + i + '_title_text': ''})

    # names on y axis - each subgroup in middle of integers - 0.5, 1.5 ,...
    subgroup_tick_dict_updated = {}
    for sub, val in subgroup_tick_dict.items():
        subgroup_tick_dict_updated[sub] = val + 0.5

    # centers axis values
    fig.update_yaxes(tickvals=list(subgroup_tick_dict_updated.values()),
                     ticktext=list(subgroup_tick_dict_updated.keys()))

    # delete y axis names [fixed] number of refs
    for i in ['', '2', '4', '5']:
        fig.update_layout({'yaxis' + i + '_title_text': ''})

    fig.update_layout({'yaxis3_title_text': 'subgroup'})

    # fixes rare bug where axis are in center and blank fields on left and right
    fig.update_xaxes(range=[min_score - 0.05, max_score + 0.05])
    fig.update_yaxes(showgrid=False, zeroline=False)

    return fig


def plot_stacked(fobject,
                 title=None,
                 other_objects=None,
                 metrics=["TPR", "PPV", "STP", "ACC", "FPR"],
                 verbose=True,
                 **kwargs):
    data = _unwrap_parity_loss_data(fobject, other_objects, metrics, verbose)

    fig = px.bar(data,
                 x='score',
                 y='label',
                 custom_data=[np.round(data.score, 3), data.metric],
                 labels={'score': 'cumulated parity loss'},
                 orientation='h',
                 color='metric',
                 color_discrete_sequence=get_default_colors(len(metrics), type='line'))
    # no outline
    fig.update_traces(marker_line_width=0)

    # hover

    if title is None:
        title = "Stacked Parity Loss Metrics"
    fig.update_layout(_fairness_theme(title))
    fig.update_yaxes(showgrid=False, zeroline=False)

    fig.update_traces(
        hovertemplate="<br>".join([
            "<b>%{label}</b><br>"
            "Metric: %{customdata[1]}<br>"
            "Parity loss: %{customdata[0]}",
            "<extra></extra>"
        ]))

    return fig


def plot_radar(fobject,
               other_objects=None,
               title=None,
               verbose=True,
               metrics=["TPR", "ACC", "PPV", "FPR", "STP"],
               **kwargs):
    data = _unwrap_parity_loss_data(fobject, other_objects, metrics, verbose)
    colors = get_default_colors(len(set(data.label)), type='line')

    fig = go.Figure()
    for i, label in enumerate(set(sorted(data.label))):
        model_data = data.loc[data.label == label, :]
        r = list(model_data.score)
        r.append(r[0])
        theta = list(model_data.metric)
        theta.append(theta[0])
        fig.add_trace(
            go.Scatterpolar(
                r=r,
                theta=theta,
                name=label,
                marker=dict(color=[colors[i] for elem in r]),
                line=dict(color=colors[i]),
                text=[label for i in r]
            )
        )

    if title is None:
        title = "Fairness Radar"
    fig.update_layout(_fairness_theme(title))

    fig.update_layout(
        polar=dict(
            radialaxis=dict(
                range=[0, max(data.score) + 0.1]
            ))
    )

    fig.update_traces(
        hovertemplate="<br>".join([
            "<b>%{text}</b><br>"
            "Metric: %{theta}<br>"
            "Parity loss: %{r:.3f}"
            "<extra></extra>"
        ]))

    return fig


def plot_performance_and_fairness(fobject,
                                  other_objects=None,
                                  fairness_metric='TPR',
                                  performance_metric='accuracy',
                                  title=None,
                                  verbose=True,
                                  **kwargs):
    data = _unwrap_parity_loss_data(fobject, other_objects, [fairness_metric], verbose)
    assert len(data.label.unique()) == len(data.label)

    performance_data = pd.DataFrame(columns=['label', 'performance_score'])
    performance_data.loc[0] = [fobject.label, _classification_performance(fobject, verbose, performance_metric)]

    if other_objects:
        other_data = pd.DataFrame()
        for i, obj in enumerate(other_objects):
            performance_data.loc[i + 1] = [obj.label, _classification_performance(obj, verbose, performance_metric)]

    data = data.merge(performance_data, on='label')

    fig = px.scatter(data,
                     x='performance_score',
                     y='score',
                     color='label',
                     custom_data=[data.label],
                     color_discrete_sequence=get_default_colors(len(data.label), 'line'))

    fig.update_traces(
        mode='markers',
        marker=dict(size=[15 for i in data.label]))

    if title is None:
        title = "Performance and Fairness"
    fig.update_layout(_fairness_theme(title))

    fig.update_yaxes(title="reversed " + fairness_metric + " parity loss",
                     autorange='reversed')
    fig.update_xaxes(title=performance_metric)

    fig.update_traces(
        hovertemplate="<br>".join([
            "<b>%{customdata[0]}</b><br>",
            fairness_metric + " parity loss: %{y:.3f}",
            performance_metric + " score: %{x:.3f}",
            "<extra></extra>"
        ]))

    return fig


def plot_heatmap(fobject,
                 other_objects=None,
                 metrics='all',
                 title=None,
                 verbose=True,
                 **kwargs):
    if metrics == 'all':
        metrics = list(fobject.parity_loss.index)

    data = _unwrap_parity_loss_data(fobject, other_objects, metrics, verbose=verbose)
    score_data = data.score.values.reshape(len(data.label.unique()), len(data.metric.unique()))

    fig = px.imshow(score_data,
                    labels=dict(x="parity loss metrics", y="model", color="Metric's parity loss"),
                    x=list(data.metric.unique()),
                    y=list(data.label.unique()),
                    color_continuous_scale=["#c7f5bf", "#8bdcbe", "#46bac2", "#4378bf", "#371ea3"])

    if title is None:
        title = "Fairness Heatmap"
    fig.update_layout(_fairness_theme(title))

    fig.update_traces(
        hovertemplate="<br>".join([
            "<b>%{y}</b><br>",
            "Parity loss metric: %{x}",
            "parity loss: %{z:.3f}",
            "<extra></extra>"
        ]))

    return fig


def plot_ceteris_paribus_cutoff(fobject,
                                other_objects=None,
                                title=None,
                                verbose=True,
                                grid_points=101,
                                subgroup=None,
                                metrics=["TPR", "ACC", "PPV", "FPR", "STP"],
                                **kwargs):
    if subgroup is None:
        raise ParameterCheckError("parameter \'subgroup\' is needed")

    protected = fobject.protected
    privileged = fobject.privileged
    if subgroup not in protected:
        raise ParameterCheckError("parameter subgroup must be in protected parameter")

    objects = [fobject]
    if other_objects is not None:
        check_other_fairness_objects(fobject, other_objects)
        for obj in other_objects:
            objects.append(obj)

    colors = get_default_colors(len(metrics), 'line')

    labels = []
    for obj in objects:
        labels.append(obj.label)

    fig = make_subplots(rows=len(objects), cols=1, subplot_titles=labels)

    for k, object in enumerate(objects):
        cutoff = object.cutoff
        y = object.y
        y_hat = object.y_hat
        data = pd.DataFrame()

        # generate data for hypothetical cutoffs
        for i in range(1, grid_points):
            cutoff[subgroup] = i / grid_points
            sub_confusion_matrix = SubgroupConfusionMatrix(y_true=y,
                                                           y_pred=y_hat,
                                                           protected=protected,
                                                           cutoff=cutoff)

            sub_confusion_matrix_metrics = SubgroupConfusionMatrixMetrics(sub_confusion_matrix)
            parity_loss = calculate_parity_loss(sub_confusion_matrix_metrics, privileged)
            parity_loss = parity_loss.loc[parity_loss.index.isin(metrics)]
            newdata = pd.DataFrame({'score': parity_loss, 'metric': parity_loss.index, 'cutoff': i / (grid_points - 1)})
            newdata = newdata.reset_index(drop=True)
            data = data.append(newdata)

        data = data.reset_index(drop=True)

        # Find minimum where NA is not present
        pivoted_data = data.pivot(index='cutoff', columns='metric', values='score')
        summed_metrics = pivoted_data.to_numpy().sum(axis=1)
        min_index = np.where(summed_metrics == min(summed_metrics))[0][0] # get first minimal index
        min_cutoff = data.cutoff.unique()[min_index]

        # make figure from individual parts
        for j, metric in enumerate(metrics):
            metric_data = data.loc[data.metric == metric, :]
            fig.add_trace(go.Scatter(x=metric_data.cutoff,
                                     y=metric_data.score,
                                     mode='lines',
                                     line=dict(color=colors[j]),
                                     name=metric,
                                     showlegend=k == 0),  # show only legend of first subplot
                          row=k + 1, col=1)

        fig.add_shape(type='line',
                      x0=min_cutoff,
                      x1=min_cutoff,
                      y0=0,
                      y1=max(data.score),
                      layer='below',
                      line=dict(color="grey",
                                dash="dot"), row=k + 1, col=1)

        fig.add_annotation(
            x=min_cutoff,
            y=max(data.score) + 0.05,
            text=f'minimum: {min_cutoff}',
            row=k + 1, col=1
        )

    fig.update_xaxes(title=f"cutoff for {subgroup}, other cutoffs constant")
    fig.update_yaxes(title="parity loss")

    if title is None:
        title = "Ceteris Paribus Cutoff"
    fig.update_layout(_fairness_theme(title))
    fig.update_layout(hovermode="x")
    fig.update_traces(
        hovertemplate="<br>".join([
            "parity loss: %{y:.3f}"
        ]))

    return fig



from dalex import Explainer
from .._plot_container import PlotContainer

class ShapleyValuesContainer(PlotContainer):
    info = {
        'name': "Shapley Values",
        'plotType': "SHAPValues",
        'plotCategory': "Observation Level",
        'requiredParams': ["model", "observation"]
    }
    options = {
        'B': { 'default': 10, 'desc': 'Number of random paths' }
    }
    def _fit(self, model, observation):
        row = observation.get_row_for_model(model)
        if row is None:
            self.set_message('Observation is not valid for given model.')
            return
        shap = model.explainer.predict_parts(
            row,
            type='shap',
            B=self.arena.get_option(self.plot_type, 'B')
        )
        intercept = shap.intercept
        result = shap.result
        result = result[result.B != 0]
        def q1(x):
            return x.quantile(0.25)
        def q3(x):
            return x.quantile(0.75)
        stats = result.groupby(['variable_name', 'variable_value']) \
            .agg({'contribution': ['mean', 'max', 'min', q1, q3]}) \
            .contribution
        stats['abs'] = stats['mean'].abs()
        stats = stats.sort_values('abs', ascending=False).reset_index()
        self.data = {
            'variables': stats.variable_name.tolist(),
            'variables_value': stats.variable_value.tolist(),
            'mean': stats['mean'].tolist(),
            'min': stats['min'].tolist(),
            'max': stats['max'].tolist(),
            'q1': stats.q1.tolist(),
            'q3': stats.q3.tolist(),
            'intercept': intercept.astype(float)
        }



import pandas as pd
import numpy as np
from pandas.api.types import is_numeric_dtype
from .._plot_container import PlotContainer

class ROCContainer(PlotContainer):
    info = {
        'name': 'Receiver Operating Characterstic',
        'plotType': 'ROC',
        'plotCategory': 'Model Performance',
        'requiredParams': ['model']
    }
    options = {
        'grid_points': { 'default': 101, 'desc': 'Maximum number of points for ROC curve' },
    }
    def _fit(self, model):
        exp = model.explainer
        if exp.model_type != 'classification':
            self.set_message('ROC plot is only available for classificators')
            return

        y_hat = exp.predict(exp.data) if exp.y_hat is None else exp.y_hat
        df = pd.DataFrame({ 'y': exp.y.astype(bool), 'y_hat': y_hat }).sort_values('y_hat', ascending=False)
        P_n = np.sum(df.y)
        N_n = df.shape[0] - P_n
        if P_n == 0 or N_n == 0:
            self.set_message('Provided dataset contains only positive or only negative cases.', 'error')
            return
        grid_points = self.arena.get_option(self.plot_type, 'grid_points')
        df['TPR'] = np.cumsum(df.y) / P_n
        df['TNR'] = 1 - (np.cumsum(1 - df.y) / N_n)
        if df.shape[0] > grid_points:
            df = df.sample(grid_points).sort_values('y_hat', ascending=False)
        self.data = {
            'cutoff': df['y_hat'].tolist(),
            'specifity': df['TNR'].tolist(),
            'sensivity': df['TPR'].tolist()
        }
    def test_arena(arena):
        if type(arena).__name__ != 'Arena' or type(arena).__module__ != 'dalex._arena.object':
            raise Exception('Invalid Arena argument')
        return next((True for model in arena.get_params('model') if model.explainer.model_type == 'classification'), False)



import numpy as np
import pandas as pd


def check_variables(variables):
    # treating variables as list simplifies code
    if variables is not None and not isinstance(variables, (str, list, np.ndarray, pd.Series)):
        raise TypeError("variables must be None or str or list or np.ndarray or pd.Series")

    if variables is None:
        variables_ = None
    elif isinstance(variables, str):
        variables_ = [variables]
    else:
        variables_ = list(variables)

    return variables_



from copy import deepcopy
import numpy as np
import pandas as pd
from dalex._explainer.helper import verbose_cat
from dalex.dataset_level._model_performance.utils import *
# -------------- Objects needed in creation of object in object.py --------------

class ConfusionMatrix:

    def __init__(self, y_true, y_pred, cutoff):
        assert len(y_true) == len(y_pred)
        assert 0 < cutoff < 1

        self.cutoff = cutoff
        self.tp = ((y_true == 1) * (y_pred >= self.cutoff)).sum()
        self.fp = ((y_true == 0) * (y_pred >= self.cutoff)).sum()
        self.tn = ((y_true == 0) * (y_pred < self.cutoff)).sum()
        self.fn = ((y_true == 1) * (y_pred < self.cutoff)).sum()


class SubgroupConfusionMatrix:

    def __init__(self, y_true, y_pred, protected, cutoff):
        assert len(y_true) == len(y_pred) == len(protected)
        assert isinstance(cutoff, dict)

        subgroups = np.unique(protected)
        sub_dict = {}

        for sub in subgroups:
            sub_indexes = np.where(protected == sub)
            sub_y_true = y_true[sub_indexes]
            sub_y_pred = y_pred[sub_indexes]

            sub_dict[sub] = ConfusionMatrix(sub_y_true, sub_y_pred, cutoff.get(sub))
        self.sub_dict = sub_dict


class SubgroupConfusionMatrixMetrics:
    """Calculate confusion matrix metrics for each subgroup

    Parameters
    -----------
    sub_confusion_matrix : SubgroupConfusionMatrix
        Object with calculated confusion matrix values for each subgroup

    Attributes
    -----------
    subgroup_confusion_matrix_metrics : dict
        Dictionary with confusion matrix metrics for each subgroup
    """

    def __init__(self, sub_confusion_matrix):
        assert isinstance(sub_confusion_matrix, SubgroupConfusionMatrix)

        matrix_dict = sub_confusion_matrix.sub_dict
        subgroup_confusion_matrix_metrics = {}

        for sub, cm in matrix_dict.items():
            tp, tn, fp, fn = cm.tp, cm.tn, cm.fp, cm.fn

            TNR = PPV = NPV = FNR = FPR = FDR = FOR = ACC = STP = np.nan

            if tp + fn > 0:
                TPR = tp / (tp + fn)
                FNR = fn / (tp + fn)
            if tn + fp > 0:
                TNR = tn / (tn + fp)
                FPR = fp / (fp + tn)
            if tp + fp > 0:
                PPV = tp / (tp + fp)
                FDR = fp / (tp + fp)
            if tn + fn > 0:
                NPV = tn / (tn + fn)
                FOR = fn / (tn + fn)
            if fp + tp + fn + tn > 0:
                ACC = (tp + tn) / (tp + tn + fp + fn)
                STP = (tp + fp) / (tp + tn + fp + fn)

            cf_metrics = {'TPR': TPR, 'TNR': TNR, 'PPV': PPV, 'NPV': NPV,
                          'FNR': FNR, 'FPR': FPR, "FDR": FDR, 'FOR': FOR, 'ACC': ACC, 'STP': STP}

            for metric in cf_metrics.keys():
                if not np.isnan(cf_metrics.get(metric)):
                    cf_metrics[metric] = round(cf_metrics.get(metric), 3)

            subgroup_confusion_matrix_metrics[sub] = deepcopy(cf_metrics)

        self.subgroup_confusion_matrix_metrics = subgroup_confusion_matrix_metrics

    def to_vertical_DataFrame(self) -> pd.DataFrame:

        columns = ['metric', 'subgroup', 'score']
        data = pd.DataFrame(columns=columns)
        metrics = self.subgroup_confusion_matrix_metrics
        for subgroup in metrics.keys():
            metric = metrics.get(subgroup)
            subgroup_vec = np.repeat(subgroup, len(metric))
            sub_df = pd.DataFrame({'metric': metric.keys(), 'subgroup': subgroup_vec, 'score': metric.values()})
            data = data.append(sub_df)
        return data

    def to_horizontal_DataFrame(self) -> pd.DataFrame:

        metrics = self.subgroup_confusion_matrix_metrics
        return pd.DataFrame(metrics).transpose()

    def __str__(self):

        return f'Object _SubgroupConfusionMatrixMetrics was converted` to data frame, top rows: \n' \
               f'{self.to_vertical_DataFrame().head().to_string()}'


# -------------- Functions needed in creation and methods of object in object.py --------------

def calculate_parity_loss(sub_confusion_matrix_metrics, privileged):
    """
    Calculates parity_loss with formula
    M_parity_loss = sum(|log(M/M_p)|)

    where
    M - vector of metrics for each subgroup
    M_p - value in metric for privileged subgroup
    """
    assert isinstance(sub_confusion_matrix_metrics, SubgroupConfusionMatrixMetrics)
    df_ratio = calculate_ratio(sub_confusion_matrix_metrics, privileged)
    columns = df_ratio.columns
    df_log = np.log(df_ratio.to_numpy())
    df_absolute_log = abs(df_log)
    df_summed = pd.DataFrame(df_absolute_log, index=df_ratio.index.values, columns=columns).apply(sum)
    return df_summed


def calculate_ratio(sub_confusion_matrix_metrics, privileged):
    """
    Calculates ratio of metrix - divides all by privileged
    Does not allow for zeros in ratios, instead puts NaN
    """
    assert isinstance(sub_confusion_matrix_metrics, SubgroupConfusionMatrixMetrics)
    df = sub_confusion_matrix_metrics.to_horizontal_DataFrame()

    privileged_index = np.where(df.index.values == privileged)[0][0]
    df_ratio = df / df.iloc[privileged_index, :]
    df_ratio = df_ratio.to_numpy()
    df_ratio[df_ratio == 0] = np.nan
    df_ratio[np.isinf(df_ratio)] = np.nan
    df_out = pd.DataFrame(df_ratio, index=df.index.values, columns=df.columns)
    return df_out




# -------------- Functions used in plots --------------
def _unwrap_parity_loss_data(fobject, other_objects, metrics, verbose):
    """ Unwrap parity loss data
    Some functions use using same computations, they are put here.
    Function creates parity loss DataFrame from fobject and stacks it with
    data from other_objects. This is private, helper function.
    """

    data = pd.DataFrame()
    data['score'] = deepcopy(fobject.parity_loss)
    data['label'] = np.repeat(fobject.label, len(fobject.parity_loss))
    data['metric'] = data.index
    data = data.reset_index(drop=True)

    if other_objects is not None:
        for object in other_objects:
            other_data = pd.DataFrame()
            other_data['score'] = deepcopy(object.parity_loss)
            other_data['label'] = np.repeat(object.label, len(object.parity_loss))
            other_data['metric'] = other_data.index
            other_data = other_data.reset_index(drop=True)
            data = data.append(other_data)

    if len(metrics) > 1:
        data = data.loc[data.metric.isin(metrics), :]
    else:
        data = data.loc[data.metric == metrics[0]]
    # checking for nans
    if any(np.isnan(data.score)):
        models_with_nans = set(data.loc[np.isnan(data.score), :].label)
        verbose_cat(f"Found NaNs in following models: {models_with_nans}", verbose)

    return data

def _fairness_theme(title):
    return {'title_text': title,
            'template': 'plotly_white',
            'title_x': 0.5,
            'title_y': 0.99,
            'titlefont': {'size': 25},
            'font': {'color': "#371ea3"},
            'margin': {'t': 78, 'b': 71, 'r': 30}}


def _metric_ratios_2_df(fobject):
    """
    Converts GroupFairnessClassification
    to elegant DataFrame with 4 columns (subgroup, metric, score, label)
    """

    data = fobject.result
    data = data.stack(dropna = False)
    data = data.reset_index()
    data.columns = ["subgroup", "metric", "score"]
    data = data.loc[data.metric.isin(["TPR", "ACC", "PPV", "FPR", "STP"])]
    data = data.loc[data.subgroup != fobject.privileged]
    data.score -= 1

    data['label'] = np.repeat(fobject.label, data.shape[0])

    return data

def _classification_performance(fobject, verbose, type='accuracy'):

    tp = tn = fp = fn = 0
    for key, val in fobject._subgroup_confusion_matrix.sub_dict.items():
        tp += val.tp
        tn += val.tn
        fp += val.fp
        fn += val.fn

    if type=='accuracy':
        return accuracy(tp, fp, tn, fn)
    if type=='auc':
        verbose_cat("Beware, that auc metric is insensitive to cutoffs", verbose)
        return auc(fobject.y_hat, fobject.y)
    if type=='recall':
        return recall(tp, fp, tn, fn)
    if type=='precision':
        return  precision(tp, fp, tn, fn)
    if type=='f1':
        return f1(tp, fp, tn, fn)
    else:
        raise TypeError(f'type \'{type}\' not supported')

# -------------- Helper functions --------------
def fairness_check_metrics():
    return ["TPR", "ACC", "PPV", "FPR", "STP"]


def intersect_unsorted(values, potential_values):
    """Return an unsorted intersection of two lists"""
    return [val for val in values if val in potential_values]


def convert_float_to_str(x, sign=""):
    """convert floats to str but do better than numpy and pandas"""
    if hasattr(x, 'values'):
        x = x.values
    y = [None]*len(x)
    for i, val in enumerate(x):
        y[i] = sign+str(val) if val > 0 else str(val)
    return y
